@@ -37,12 +37,13 @@ may be strictly algorithmic, may interact with human subjects, or may be embedde
 sites.
 """
 
-__version__ = '1.0.6'
+__version__ = "1.0.9"
 
 import collections
 import collections.abc as abc
 import math
 import numbers
+import operator
 import pylru
 import random
 import sys
@@ -111,7 +112,7 @@ class Memory(dict):
         self._learning_time_increment = learning_time_increment
         self._retrieval_time_increment = retrieval_time_increment
         self._activation_history = None
-        self.reset(bool(optimized_learning))
+        self.reset(optimized_learning=bool(optimized_learning))
 
     def __repr__(self):
         return f"<Memory {dict(self.values())}>"
@@ -119,8 +120,11 @@ class Memory(dict):
     def __str__(self):
         return f"<Memory {id(self)}>"
 
-    def reset(self, optimized_learning=None):
-        """Deletes all the Memory's chunks and resets its time to zero.
+    def reset(self, preserve_prepopulated=False, optimized_learning=None):
+        """Deletes the Memory's chunks and resets its time to zero.
+        If *preserve_prepopulated* is false it deletes all chunks; if it is true it
+        deletes all chunk references later than time zero, completely deleting those
+        chunks that were created at time greater than zero.
         If *optimized_learning* is not None it sets the Memory's :attr:`optimized_learning`
         parameter; otherwise it leaves it unchanged. This Memory's :attr:`noise`,
         :attr:`decay`, :attr:`temperature`, :attr:`threshold` and :attr:`mismatch`
@@ -128,10 +132,17 @@ class Memory(dict):
         """
         if optimized_learning and self._decay >= 1:
             raise RuntimeError(f"Optimized learning cannot be enabled if the decay, {self._decay}, is not less than 1")
+        if preserve_prepopulated:
+            preserved = {k: v for k, v in self.items() if v._creation == 0}
+            print(preserved)
         self.clear()
         self._time = 0
         if optimized_learning is not None:
             self._optimized_learning = bool(optimized_learning)
+        if preserve_prepopulated:
+            for k, v in preserved.items():
+                v._references = 0 if self._optimized_learning else [0]
+                self[k] = v
 
     @property
     def time(self):
@@ -341,12 +352,8 @@ class Memory(dict):
         >>> m = Memory()
         >>> m.learn(color="red", size=3)
         True
-        >>> m.advance()
-        1
         >>> m.learn(color="red", size=5)
         True
-        >>> m.advance()
-        2
         >>> m.activation_history = []
         >>> m.blend("size", color="red")
         4.483378114110536
@@ -400,8 +407,9 @@ class Memory(dict):
     _similarity_cache = pylru.lrucache(SIMILARITY_CACHE_SIZE)
 
     def _similarity(self, x, y, attribute):
+        # always returns the "natural" similarity
         if x == y:
-            return 0 if Memory._use_actr_similarity else 1
+            return 1
         fn = self._similarity_functions.get(attribute)
         if fn:
             signature = (x, y, attribute)
@@ -418,10 +426,12 @@ class Memory(dict):
             elif result > Memory._maximum_similarity:
                 warn(f"similarity value is greater than the maximum allowed, {Memory._maximum_similarity}, so that maximum value is being used instead")
                 result = Memory._maximum_similarity
+            if Memory._use_actr_similarity:
+                result += 1
             self._similarity_cache[signature] = result
             return result
         else:
-            return -1 if Memory._use_actr_similarity else 0
+            return 0
 
     def learn(self, advance=None, **kwargs):
         """Adds, or reinforces, a chunk in this Memory with the attributes specified by *kwargs*.
@@ -452,8 +462,6 @@ class Memory(dict):
         True
         >>> m.learn(color="red", size=4)
         False
-        >>> m.advance()
-        1
         >>> m.retrieve(color="red")
         <Chunk 0000 {'color': 'red', 'size': 4}>
 
@@ -521,8 +529,8 @@ class Memory(dict):
         of match.
 
         Before performing the retrieval :meth:`advance` is called with the value of
-        *advance* as its argument. If *advance* is not supplied the current value of
-        attr:`retrieval_time_increment` is used; unless changed by the programmer this
+        *advance* as its argument. If *advance* is not supplied the current value
+        of :attr:`retrieval_time_increment` is used; unless changed by the programmer this
         default value is zero. The advance of time does not occur if an error is raised
         when attempting to perform the retrieval.
 
@@ -535,12 +543,8 @@ class Memory(dict):
         >>> m = Memory()
         >>> m.learn(widget="thromdibulator", color="red", size=2)
         True
-        >>> m.advance()
-        1
         >>> m.learn(widget="snackleizer", color="blue", size=1)
         True
-        >>> m.advance()
-        2
         >>> m.retrieve(color="blue")["widget"]
         'snackleizer'
         """
@@ -598,8 +602,9 @@ class Memory(dict):
                 if self._conditions.keys() <= chunk.keys(): # subset
                     if self._memory._mismatch is not None:
                         activation = chunk._activation(True)
-                        mismatch = self._memory._mismatch * sum(self._memory._similarity(c, chunk[s], s)
-                                                                for s, c in self._conditions.items())
+                        mismatch = (self._memory._mismatch
+                                    * sum(self._memory._similarity(c, chunk[s], s) - 1
+                                          for s, c in self._conditions.items()))
                         total = activation + mismatch
                         if self._memory._activation_history is not None:
                             history = self._memory._activation_history[-1]
@@ -618,13 +623,15 @@ class Memory(dict):
         return self._Activations(self, conditions)
 
     def _partial_match(self, conditions):
-        best_chunk = None
+        best_chunks = []
         best_activation = self._threshold
         for chunk, activation in self._activations(conditions):
-            if activation >= self._threshold:
-                best_chunk = chunk
+            if activation > best_activation:
+                best_chunks = [chunk]
                 best_activation = activation
-        return best_chunk
+            elif activation == best_activation:
+                best_chunks.add(chunk)
+        return random.choice(best_chunks) if best_chunks else None
 
     def blend(self, outcome_attribute, advance=None, **kwargs):
         """Returns a blended value for the given attribute of those chunks matching *kwargs*, and which contains *outcome_attribute*.
@@ -632,28 +639,21 @@ class Memory(dict):
         *outcome_attribute*. If any matching chunk has a value of *outcome_attribute*
         value that is not a real number a :exc:`TypeError` is raised.
 
-        Before performing the blending operation  :meth:`advance` is called with the value
-        of *advance* as its argument. If *advance* is not supplied the current value of
-        attr:`retrieval_time_increment` is used; unless changed by the programmer this
+        Before performing the blending operation :meth:`advance` is called with the value
+        of *advance* as its argument. If *advance* is not supplied the current value
+        of :attr:`retrieval_time_increment` is used; unless changed by the programmer this
         default value is zero. The advance of time does not occur if an error is raised
         when attempting to perform the blending operation.
 
         >>> m = Memory()
         >>> m.learn(color="red", size=2)
         True
-        >>> m.advance()
-        1
         >>> m.learn(color="blue", size=30)
         True
-        >>> m.advance()
-        2
         >>> m.learn(color="red", size=1)
         True
-        >>> m.advance()
-        3
         >>> m.blend("size", color="red")
         1.1548387620911693
-
         """
         old = self._advance(advance, self._retrieval_time_increment)
         try:
@@ -687,6 +687,82 @@ class Memory(dict):
                 # finish normally
                 self._time = old
 
+    def best_blend(self, outcome_attribute, iterable, select_attribute=None, advance=None, minimize=False):
+        """Returns two values (as a 2-tuple), describing the extreme blended value of the *outcome_attribute* over the values provided by *iterable*.
+        The extreme value is normally the maximum, but can be made the minimum by setting
+        *minimize* to True. The values returned by *iterable* should be dictionary-like
+        objects that can be passed as the *kwargs* argument to :meth:`blend`. The first
+        return value is the *kwargs* value producing the best blended value, and second is
+        that blended value. If there is a tie, with two or more *kwargs* values all
+        producing the same, best blended value, then one of them is chosen randomly. If
+        none of the values from *iterable* result in blended values of *outcome_attribute*
+        then both return values are ``None``.
+
+        This operation is particularly useful for building Instance Based Learning models.
+
+        For the common case where *iterable* iterates over only the values of a single
+        slot the *select_attribute* parameter may be used to simplify the iteration. If
+        *select_attribute* is supplied and is not ``None`` then *iterable* should produce
+        values of that slot instead of dictionary-like objects. Similarly the first return
+        value will be the slot value rather than a dictionary-like object. The end of the
+        example below demonstrates this.
+
+        Before performing the blending operations :meth:`advance` is called, only once,
+        with the value of *advance* as its argument. If *advance* is not supplied the
+        current value of :attr:`retrieval_time_increment` is used; unless changed by the
+        programmer this default value is zero. The advance of time does not occur if an
+        error is raised when attempting to perform a blending operation.
+
+        >>> m = Memory()
+        >>> m.learn(color="red", utility=1)
+        True
+        >>> m.learn(color="blue", utility=2)
+        True
+        >>> m.learn(color="red", utility=1.8)
+        True
+        >>> m.learn(color="blue", utility=0.9)
+        True
+        >>> m.best_blend("utility", ({"color": c} for c in ("red", "blue")))
+        ({'color': 'blue'}, 1.4868)
+        >>> m.learn(color="blue", utility=-1)
+        True
+        >>> m.best_blend("utility", ("red", "blue"), "color")
+        ('red', 1.2127)
+
+        """
+        comparator = operator.gt if not minimize else operator.lt
+        old = self._advance(advance, self._retrieval_time_increment)
+        try:
+            best_value = -math.inf if not minimize else math.inf
+            best_args = []
+            for thing in iterable:
+                if select_attribute is not None:
+                    kwargs = { select_attribute : thing }
+                else:
+                    kwargs = thing
+                value = self.blend(outcome_attribute, advance=0, **kwargs)
+                if value is None:
+                    pass
+                elif value == best_value:
+                    best_args.append(kwargs)
+                elif comparator(value, best_value):
+                    best_args = [ kwargs ]
+                    best_value = value
+            old = None
+            if best_args:
+                result = random.choice(best_args)
+                if select_attribute is not None:
+                    result = result[select_attribute]
+                return result, best_value
+            else:
+                return None, None
+        finally:
+            if old is not None:
+                # Don't advance if there's an error or for some other reason we don't
+                # finish normally
+                self._time = old
+
+
 def use_actr_similarity(value=None):
     """Whether to use "natural" similarity values, or traditional ACT-R ones.
     PyACTUp normally uses a "natural" representation of similarities, where two values
